"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable operator-linebreak */
const abap_adt_api_1 = require("abap-adt-api");
const cli_ux_1 = require("cli-ux");
exports.PACKAGE = "DEVC/K";
const PACKAGETYPE = { desc: "Package", type: exports.PACKAGE };
exports.treatableTypes = [
    { desc: "Class", type: "CLAS/OC" },
    { desc: "FunctionFroup", type: "FUGR/F" },
    { desc: "Interface", type: "INTF/OI" },
    { desc: "Include", type: "PROG/I" },
    { desc: "Program", type: "PROG/P" }
];
exports.supportedTypes = [PACKAGETYPE, ...exports.treatableTypes];
const extensions = new Map([
    ["CLAS/I", ".clas"],
    ["FUGR/FF", ".fugr"],
    ["FUGR/I", ".prog"],
    ["INTF/OI", ".intf"],
    ["PROG/I", ".prog"],
    ["PROG/P", ".prog"]
]);
exports.includeName = (i) => `${i.name}${extensions.get(i.type) || ""}.abap`;
const follow = (base, next) => {
    if (next.startsWith("/"))
        return next;
    if (next.startsWith("./")) {
        const rest = next.substr(2);
        return base.endsWith("/")
            ? `${base}${rest}`
            : `${base.replace(/\/[^/]*$/, "/")}${rest}`;
    }
    return `${base.replace(/\$/, "")}/${next}`;
};
const fromObject = (o, url) => {
    const { type, name, url: metaUrl } = o;
    return { type, name, metaUrl, sourceUrl: follow(metaUrl, url) };
};
const canProcess = (n) => n.OBJECT_URI &&
    n.OBJECT_NAME &&
    exports.treatableTypes.find(t => t.type === n.OBJECT_TYPE);
exports.supportedType = (type) => !!exports.supportedTypes.find(t => t.type === type);
const fromNode = (n) => ({
    type: n.OBJECT_TYPE,
    name: n.OBJECT_NAME,
    url: n.OBJECT_URI
});
async function expandPackage(client, name, notifier, recursive) {
    const objects = [];
    notifier(`Expanding package ${name}`);
    const { nodes } = await client.statelessClone.nodeContents(exports.PACKAGE, name);
    const children = nodes.filter(canProcess).map(fromNode);
    objects.push(...children);
    if (recursive)
        for (const child of nodes.filter(n => n.OBJECT_TYPE === exports.PACKAGE && n.OBJECT_NAME)) {
            const granchildren = await expandPackage(client, child.OBJECT_NAME, notifier, recursive);
            objects.push(...granchildren);
        }
    return objects;
}
async function list(client, type, name, notifier, recursive = false) {
    type = type === null || type === void 0 ? void 0 : type.toUpperCase();
    name = name === null || name === void 0 ? void 0 : name.toUpperCase();
    if (!exports.supportedType(type))
        cli_ux_1.cli.error(`Type ${type} is not supported`);
    const list = await client.statelessClone.searchObject(name, type);
    const found = list.find(c => c["adtcore:type"] === type && c["adtcore:name"] === name);
    if (!found)
        cli_ux_1.cli.error(`Object ${type} ${name} not found`);
    switch (type) {
        case exports.PACKAGE:
            return expandPackage(client, name, notifier, recursive);
        default:
            return [{ type, name, url: found["adtcore:uri"] }];
    }
}
exports.list = list;
async function expandClass(client, base) {
    const stru = await client.statelessClone.objectStructure(base.url);
    if (abap_adt_api_1.isClassStructure(stru))
        return stru.includes.map(i => {
            var _a;
            const url = ((_a = i.links.find(l => l.rel === "http://www.sap.com/adt/relations/source" &&
                l.type === "text/plain")) === null || _a === void 0 ? void 0 : _a.href) || i["abapsource:sourceUri"];
            return {
                name: `${base.name} ${i["class:includeType"]}`,
                type: i["adtcore:type"],
                metaUrl: base.url,
                sourceUrl: follow(base.url, url)
            };
        });
    return [];
}
async function expandFunc(client, name, metaUrl) {
    var _a;
    const { nodes } = await client.statelessClone.nodeContents("FUGR/F", name);
    const [ns = "", base = ""] = ((_a = name.match(/(\/[^/]+\/)?(.*)/)) === null || _a === void 0 ? void 0 : _a.slice(1)) || [];
    const prefix = `${ns}L${base}`;
    const goodInclude = (n) => n.startsWith(prefix) && !n.endsWith("UXX");
    const filtFunc = (n) => (n.OBJECT_NAME && n.OBJECT_TYPE === "FUGR/FF") ||
        (n.OBJECT_TYPE === "FUGR/I" && goodInclude(n.OBJECT_NAME));
    const children = [];
    for (const n of nodes.filter(filtFunc)) {
        const base = n.OBJECT_URI.replace(/\/source\/main/i, "");
        const structure = await client.statelessClone.objectStructure(base);
        const sourceUrl = abap_adt_api_1.ADTClient.mainInclude(structure, true);
        children.push({
            type: n.OBJECT_TYPE,
            name: n.OBJECT_NAME,
            sourceUrl,
            metaUrl
        });
    }
    return children;
}
async function expand(client, base) {
    switch (base.type) {
        case "CLAS/OC":
            return expandClass(client, base);
        case "FUGR/F":
            return expandFunc(client, base.name, base.url);
    }
    const structure = await client.statelessClone.objectStructure(base.url);
    const url = abap_adt_api_1.ADTClient.mainInclude(structure, true);
    return [fromObject(base, url)];
}
exports.expand = expand;
